import pandas as pd
import scanpy as sc
import os
import matplotlib.pyplot as plt

def load_data(sc_file_path,anno_files,bulk_file_path):
    print("""Loading data from the specified path, please wait...""")
    sc_file_path = [sc_file_path]
    data_list = []
    for file in sc_file_path:
        data = pd.read_csv(file, index_col=0)
        data_list.append(data)
    data = pd.concat(data_list, axis=1)
    num_rows, num_columns = data.shape
    print(f"Loaded data from {file}")
    print(f"Gene number: {num_rows}")
    print(f"Cell number: {num_columns}")
    anno_list = []
    anno_files = [anno_files]
    for file in anno_files:
        anno = pd.read_csv(file)
        anno_list.append(anno)
    anno = pd.concat(anno_list)
    bulk_data = pd.read_csv(bulk_file_path, index_col=0)
    num_rows, num_columns = bulk_data.shape
    print("Loaded bulk data:")
    print(f"Gene number:{num_rows}")
    print(f"Sample number:{num_columns}")
    print("""Successfully loaded data from the specified path""")

    return data,anno,bulk_data


def filter_genes(bulk_data, sc_data, min_cells=20):
    print('Genes being screened Waiting...')
    common_genes = list(set(bulk_data.index).intersection(set(sc_data.index)))
    bulk_data = bulk_data.loc[common_genes]
    sc_data = sc_data.loc[common_genes]

    expr_counts = sc_data.astype(bool).sum(axis=1)
    expr_genes = expr_counts[expr_counts >= min_cells].index
    bulk_data = bulk_data.loc[expr_genes]
    sc_data = sc_data.loc[expr_genes]

    print(f'Genes retained as expressed in more than {min_cells} cells：{len(expr_genes)} genes')

    return bulk_data, sc_data

def preprocess_and_plot_umap(adata, anno, save_path=None):
    """
    Preprocessing of Annddata objects and generation of umap images。

    Parameters:
    - adata (AnnData): Annddata objects containing single-cell data
    - anno (DataFrame): Dataframe containing cellular annotation information
    - save_path (str, optional): Path and file name where umap images were saved. If not provided, save under the current path by default.
    """

    # Conversion to anddata
    adata = sc.AnnData(X=adata.T)

    # Adding cell annotation information to annddata
    adata.obs['celltype'] = ''
    for idx, row in anno.iterrows():
        adata.obs.loc[row['cellname'], 'celltype'] = row['celltype']

    # pre-processing
    print("Started preprocessing the data...")
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.highly_variable_genes(adata, n_top_genes=2000)
    sc.pp.scale(adata, max_value=10)
    sc.pp.pca(adata, n_comps=50, use_highly_variable=True)
    sc.pp.neighbors(adata)
    sc.tl.umap(adata)
    print("Preprocessing complete...")

    # Save UAMP images
    if save_path is None:
        save_path = os.path.join(os.getcwd(), "Orig_UMAP.png")
    else:
        save_path = os.path.join(os.getcwd(), save_path)

    print(f"The umap image will be saved to: {save_path}")
    sc.pl.umap(adata, color=['celltype'], legend_loc='on data', title='', frameon=True, show=False)
    plt.savefig(save_path)
    plt.close()

    # Returns preprocessed Annddata objects
    return adata






import os
import scanpy as sc
import pandas as pd
import numpy as np
import warnings
import matplotlib.pyplot as plt
from sklearn.preprocessing import scale
from tqdm import tqdm
from scipy.optimize import nnls
import anndata as ad
def scale_data(exp):
    """
    Parameter:
    - exp (pd.DataFrame): FPKM Expression

    Return:
    - scaled_matrix (pd.DataFrame)
    """

    # Expression data were log10 transformed
    exp_log = np.log10(exp + 1)

    # Standardization
    print("Starting with standardized data...")
    scaled_matrix = exp_log.apply(lambda x: scale(x.values), axis=1, result_type='expand')
    scaled_matrix[np.isnan(scaled_matrix)] = 0

    # Set the names of the rows and columns
    scaled_matrix.index = exp.index
    scaled_matrix.columns = exp.columns

    print("Data normalization is complete!")

    return scaled_matrix

def compute_cell_type_proportions(adata, bulk_data, num_top_genes=50):
    """
    Specific expression genes were solved based on each cell type, and the proportion of each cell type in the sample was calculated.

    Parameter:
    - adata (AnnData): Annddata objects containing single-cell data
    - bulk_data (DataFrame): Dataframe containing bulk data, with each column representing one sample
    - num_top_genes (int, optional): Number of specific genes selected per cell type, default 50

    Return:
    - cell_type_proportions_df (DataFrame): The proportion of each cell type in each sample, with the sample name as the row and the cell type as the column of the dataframe
    """

    # Specifically expressed genes were solved based on each cell type
    print('Solve for each cell type specific expression gene...')
    sc.tl.rank_genes_groups(adata, groupby='celltype', method='t-test')
    all_genes = []
    for ct in adata.obs['celltype'].unique():
        top_genes = adata.uns['rank_genes_groups']['names'][ct][:num_top_genes]
        all_genes.extend(top_genes)

    # Screen for specific expression genes of each cell type, retaining genes that are only specific to that cell type and not specific to other types of cells
    specific_genes = np.unique(all_genes)
    num_specific_genes = len(specific_genes)
    print(f"Retained {num_specific_genes} specific genes")
    # Obtain the gene expression matrix of the reference cell type from the AnaData object
    ref_data = adata[:, specific_genes].X.T
    ref_cell_types = adata.obs['celltype'].cat.categories.values
    ref_cell_types_data = np.zeros((len(specific_genes), len(ref_cell_types)))
    for i, cell_type in enumerate(ref_cell_types):
        cell_type_data = adata[adata.obs['celltype'] == cell_type, specific_genes].X.mean(axis=0)
        ref_cell_types_data[:, i] = cell_type_data

    # Initialize the DataFrame that saves the proportion of cell types
    cell_type_proportions_df = pd.DataFrame(index=bulk_data.columns, columns=ref_cell_types)

    # Cycle to calculate the proportion of cell types for each sample
    for col in bulk_data.columns:
        bulk_gene_counts = bulk_data.loc[specific_genes][col].values

        # NNLS deconvolution of reference gene expression data
        cell_type_proportions, _ = nnls(ref_cell_types_data, bulk_gene_counts)

        # Save cell type ratio to DataFrame
        cell_type_proportions_df.loc[col] = cell_type_proportions

    # Normalize the cell proportion data so that the sum of the proportions for each sample is one
    cell_type_proportions_df = cell_type_proportions_df.div(cell_type_proportions_df.sum(axis=1), axis=0)
    visualize_cell_type_proportions(cell_type_proportions_df.astype(float))
    return cell_type_proportions_df

def visualize_cell_type_proportions(cell_type_proportions_df):
    """
    Visualize the proportion of each cell type in each sample

    Parameters:
    - cell_type_proportions_df (DataFrame): The proportion of each cell type in each sample, DataFrame with sample name as row and cell type as column
    """

    # Set Drawing Style
    plt.style.use('seaborn-dark')

    # Obtain the number of samples and cell types
    num_samples, num_cell_types = cell_type_proportions_df.shape

    # Set Color Mapping
    colors = plt.cm.get_cmap('tab20c', num_cell_types)

    # Create drawings and axes
    fig, ax = plt.subplots(figsize=(10, 6))

    # Draw Stacked bar chart
    bottom = np.zeros(num_samples)
    for i, cell_type in enumerate(cell_type_proportions_df.columns):
        proportions = cell_type_proportions_df[cell_type]
        ax.bar(range(num_samples), proportions, bottom=bottom, color=colors(i), label=cell_type)
        bottom += proportions

    # Set graphic titles and labels
    ax.set_title("Cell Type Proportions")
    ax.set_xlabel("Samples")
    ax.set_ylabel("Proportions")

    # add legend
    ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left', borderaxespad=0.)

    # Automatically adjust subgraph layout
    fig.tight_layout()

    # display graphics
    plt.show()



from tkinter import filedialog
import matplotlib.pyplot as plt
from PIL import Image
from IPython.display import display
import ipywidgets as widgets
import os
import scanpy as sc
import pandas as pd
from combat.pycombat import pycombat
import numpy as np
import warnings
import matplotlib.pyplot as plt
from sklearn.preprocessing import scale
from tqdm import tqdm
from scipy.optimize import nnls
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from scipy.stats import spearmanr
import tensorflow as tf
from tensorflow.keras import layers
import random
import anndata as ad
import argparse



def loop_choose_std(best_model, cell_type_proportions_df, common_genes_metabolism, data_metabolism, anno,
                            num_samples=3000, latent_dim=64, z_std=1, save_path=None):

    z_std_list = [(z_std/2),(z_std),(z_std+z_std/2)]
    img = []
    for z_std in z_std_list:
        z_mean = np.zeros(latent_dim)
        z = np.random.normal(loc=z_mean, scale=z_std, size=(num_samples, latent_dim))
        # Calculate the number of each cell type
        cell_type_counts = anno.groupby('celltype').size()
        cell_type_proportions_sc = cell_type_counts / cell_type_counts.sum()
        cell_type_proportions_df_sc = cell_type_proportions_sc.to_frame(name='proportions').reset_index()
        cell_type_proportions_df_sc = cell_type_proportions_df_sc.set_index('celltype')
        cell_type_proportions_df_transposed_sc = cell_type_proportions_df_sc.T
        cell_type_proportions_df_transposed_sc.reset_index(drop=True, inplace=True)
        cell_types_list_sc = cell_type_proportions_df_transposed_sc.columns
        cell_types = np.random.choice(cell_types_list_sc, size=num_samples, p=cell_type_proportions_sc.values)
        # Encoding cell types with OneHotEncoder
        encoder = OneHotEncoder(sparse=False)
        c = encoder.fit_transform(cell_types.reshape(-1, 1))
        # Merge latent variables and conditions into a tensor
        z_c = np.hstack([z, c])
        # Input CVAE model to generate sample data and cell type annotations
        x_pred, y_pred = best_model.decoder.predict(z_c), np.argmax(c, axis=1)
        print(f"Cell generation completed, total generation: {num_samples} cells")
        # Save the generated data and labels as DataFrame
        df = pd.DataFrame(x_pred)
        df['celltype'] = y_pred
        labels = cell_type_proportions_df.columns.tolist()
        label_dict = {i: label for i, label in enumerate(labels)}
        df['celltype'] = df['celltype'].replace(label_dict)
        generated_data = pd.DataFrame(x_pred, columns=common_genes_metabolism)
        print("Continue processing the generated data")
        value_proportions = data_metabolism.T.apply(lambda x: x.value_counts(normalize=True))
        for col in generated_data.columns:
        # Filter out data with a ratio of 0 and sort in descending order
            proportions = value_proportions[col][value_proportions[col] > 0.0001].sort_index(ascending=False)
            values = proportions.index.values
            cum_proportions = np.cumsum(proportions.values)

        # Calculate the number of elements that should be assigned to each value based on proportion
            num_elements_per_value = (cum_proportions * generated_data.shape[0]).astype(int)
            num_elements_diff = np.diff(np.concatenate(([0], num_elements_per_value)))

        # Correct the number of the last element to match the number
            num_elements_diff[-1] = generated_data.shape[0] - np.sum(num_elements_diff[:-1])

        # Generate a list of all values based on the number of elements and arrange them in descending order
            all_values = np.repeat(values, num_elements_diff)

        # For generated_ Sort the current column of data in descending order and obtain the sorted index
            sorted_indices = np.argsort(-generated_data.loc[:, col])

        # Assign the arranged values to generated_ Data, keep row order unchanged
            generated_data.loc[sorted_indices, col] = all_values
        print( "Processing completed, conducting simulation data quality analysis. Please try different standard deviation values based on the UMAP clustering diagram")
        adata = ad.AnnData(X=generated_data, obs={'cell_type': df['celltype'].values})
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, n_top_genes=1000)
        sc.pp.scale(adata)
        sc.tl.pca(adata, svd_solver='arpack')
        sc.pp.neighbors(adata, n_neighbors=15, n_pcs=50)
        sc.tl.umap(adata)
        save_path = os.path.join(os.getcwd(), "Simulation_UMAP_Metabolism_std_" + str(z_std) + ".png")
        print(f"Total generated: {num_samples} cells")
        print(f"Metabolism_ UMAP image will be saved to: {save_path}")
        print(f"The current standard deviation value is: {z_std},Please adjust according to the UMAP clustering effect")
        std_str = str(z_std)
        sc.pl.umap(adata, color=['cell_type'], legend_loc='on data',title='The current std is'+std_str,frameon=True, show=False)
        img.append(save_path)
        plt.savefig(save_path)
        plt.close()

    return img,z_std_list


def select_image(img_list):
    images = []
    for img_path in img_list:
        image = Image.open(img_path)
        images.append(image)

    fig, axs = plt.subplots(1, len(images), figsize=(10, 6))  # 调整子图大小
    fig.tight_layout()

    for i, image in enumerate(images):
        axs[i].imshow(image)
        axs[i].axis('off')

    plt.show()

    selected_index = None
    selected_type = None
    while selected_index is None:
        user_input = input("Please select the image serial number (input serial number) or enter a new standard deviation (input s) for the graph: ")
        if user_input == "s":
            new_std = float(input("Please enter a new standard deviation value: "))
            selected_index = new_std
            selected_type = "std"
        else:
            try:
                selected_index = int(user_input) - 1
                if selected_index < 0 or selected_index >= len(images):
                    print("Invalid image sequence number, please re-enter")
                    selected_index = None
            except ValueError:
                print("Invalid input, please re-enter")

    if selected_type == "std":
        print(f"The new standard deviation you have selected is {selected_index}")
        return selected_index, selected_type
    else:
        print(f"The serial number of the image you selected is {selected_index + 1}")
        return selected_index, selected_type



def simulate_bulk_data(cell_type_proportions_df, df, bulk_metabolism_ratio,common_genes_metabolism, num_simulations=20, num_cells=5000 ):
    selected_gene_expression_df_total = []
    selected_celltype = []
    selected_samplename = []

    for sample in range(cell_type_proportions_df.shape[0]):
        cell_type_proportions_df_tem = pd.DataFrame(cell_type_proportions_df.iloc[sample, :]).T
        col_sum = cell_type_proportions_df_tem.sum(axis=0)
        zero_cols = col_sum[col_sum == 0].index.tolist()
        cell_type_proportions_df_tem = cell_type_proportions_df_tem.loc[:,
                                       ~cell_type_proportions_df_tem.columns.isin(zero_cols)]
        print(f"Simulate {cell_type_proportions_df.index[sample]} {num_simulations} times")
        min_distance = float('inf')
        selected_gene_expression_df = None
        bulk_metabolism_ratio_tem = bulk_metabolism_ratio.iloc[:, sample]
        for i in range(num_simulations):
            cells_list = []
            seed = random.randint(1, 1000)

            for j in range(cell_type_proportions_df_tem.shape[1]):
                n_cells = (cell_type_proportions_df_tem.iloc[:, j] * num_cells).astype(int)
                cells = df[df['celltype'] == cell_type_proportions_df_tem.columns[j]].sample(n=n_cells.values)
                cells_list.append(cells)

            cells_df = pd.concat(cells_list)
            cells_label = cells_df.iloc[:, -1]
            cells_df = cells_df.iloc[:, :-1]

            gene_expression_df = cells_df.sum(axis=0)
            corr, pval = spearmanr(gene_expression_df, bulk_metabolism_ratio_tem)
            print(f"Round {i + 1} Spearman Correlation: {corr.mean():.5f}")

            gene_expression_df = gene_expression_df / gene_expression_df.sum()
            distance = np.linalg.norm(gene_expression_df.values - bulk_metabolism_ratio_tem.values)

            if distance < min_distance:
                min_distance = distance
                selected_gene_expression_df = cells_df
                selected_cells_label = cells_label
                selected_sample_name = pd.DataFrame([cell_type_proportions_df.index[sample]] * len(cells_df))
                Item = i
                best_corr = corr

            print(f"Round {i + 1} Euclidean Distance: {distance:.5f}")

        plt.scatter(selected_gene_expression_df.sum(axis=0).rank(), bulk_metabolism_ratio_tem.rank(), alpha=0.5)
        save_path = os.path.join(os.getcwd(), f"{cell_type_proportions_df.index[sample]}_corr.png")
        plt.savefig(save_path)
        plt.close()
        print(f"Save the correlation scatter plot to:{save_path}")

        selected_gene_expression_df_total.append(selected_gene_expression_df)
        selected_celltype.append(selected_cells_label)
        selected_samplename.append(selected_sample_name)
        print(
            f"Select the fitting result of the {Item + 1}th order,with a sample distance of {min_distance},and a correlation of{best_corr}")

    selected_gene_expression_df_total = pd.concat(selected_gene_expression_df_total, axis=0)
    selected_gene_expression_df_total.columns = common_genes_metabolism
    selected_gene_expression_df_total.index = ['Cell' + str(i) for i in
                                               range(1, selected_gene_expression_df_total.shape[0] + 1)]

    selected_celltype = pd.concat(selected_celltype, axis=0)
    selected_celltype = pd.DataFrame(selected_celltype)
    selected_celltype.index = ['Cell' + str(i) for i in range(1, selected_celltype.shape[0] + 1)]

    selected_samplename = pd.concat(selected_samplename, axis=0)
    selected_samplename.index = ['Cell' + str(i) for i in range(1, selected_samplename.shape[0] + 1)]

    return selected_gene_expression_df_total, selected_celltype, selected_samplename

def generate_umap_plots(selected_gene_expression_df_total, selected_celltype, selected_samplename, save_path=None):
    adata = ad.AnnData(X=selected_gene_expression_df_total)
    adata.obs['celltype'] = selected_celltype['celltype']
    adata.obs['sample'] = selected_samplename.iloc[:, 0]
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.highly_variable_genes(adata, n_top_genes=1000)
    sc.pp.scale(adata)
    sc.tl.pca(adata, svd_solver='arpack')
    sc.pp.neighbors(adata, n_neighbors=15, n_pcs=50)
    sc.tl.umap(adata)

    # Save UMAP plots
    if save_path is None:
        save_path_celltype = os.path.join(os.getcwd(), "Bulk_Simulation_UMAP_celltype.png")
        save_path_sample = os.path.join(os.getcwd(), "Bulk_Simulation_UMAP_sample.png")
    else:
        save_path_celltype = os.path.join(save_path, "Bulk_Simulation_UMAP_celltype.png")
        save_path_sample = os.path.join(save_path, "Bulk_Simulation_UMAP_sample.png")

    print(f"Total generated cells: {selected_gene_expression_df_total.shape[0]}")
    print(f"UMAP plots will be saved to: {save_path}")

    sc.pl.umap(adata, color=['celltype'], legend_loc='on data', title='', frameon=True, show=False)
    plt.savefig(save_path_celltype)
    plt.close()
    sc.pl.umap(adata, color=['celltype'], legend_loc='on data', title='', frameon=True)
    sc.pl.umap(adata, color=['sample'], legend_loc='on data', title='', frameon=True, show=False)
    plt.savefig(save_path_sample)
    plt.close()
    sc.pl.umap(adata, color=['sample'], legend_loc='on data', title='', frameon=True)
    return adata